@@ -4,6 +4,7 @@ import (
 	"fmt"
 	"os/exec"
 	"strings"
+	"time"
 
 	"github.com/docker/docker/pkg/integration/checker"
 	"github.com/docker/docker/pkg/stringid"
@@ -11,7 +12,6 @@ import (
 )
 
 func (s *DockerSuite) TestRmiWithContainerFails(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	errSubstr := "is using it"
 
 	// create a container
@@ -33,7 +33,6 @@ func (s *DockerSuite) TestRmiWithContainerFails(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiTag(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	imagesBefore, _ := dockerCmd(c, "images", "-a")
 	dockerCmd(c, "tag", "busybox", "utest:tag1")
 	dockerCmd(c, "tag", "busybox", "utest/docker:tag2")
@@ -62,10 +61,17 @@ func (s *DockerSuite) TestRmiTag(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiImgIDMultipleTag(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	out, _ := dockerCmd(c, "run", "-d", "busybox", "/bin/sh", "-c", "mkdir '/busybox-one'")
 
 	containerID := strings.TrimSpace(out)
+
+	// Wait for it to exit as cannot commit a running container on Windows, and
+	// it will take a few seconds to exit
+	if daemonPlatform == "windows" {
+		err := waitExited(containerID, 60*time.Second)
+		c.Assert(err, check.IsNil)
+	}
+
 	dockerCmd(c, "commit", containerID, "busybox-one")
 
 	imagesBefore, _ := dockerCmd(c, "images", "-a")
@@ -76,16 +82,15 @@ func (s *DockerSuite) TestRmiImgIDMultipleTag(c *check.C) {
 	// tag busybox to create 2 more images with same imageID
 	c.Assert(strings.Count(imagesAfter, "\n"), checker.Equals, strings.Count(imagesBefore, "\n")+2, check.Commentf("docker images shows: %q\n", imagesAfter))
 
-	imgID, err := inspectField("busybox-one:tag1", "Id")
-	c.Assert(err, checker.IsNil)
+	imgID := inspectField(c, "busybox-one:tag1", "Id")
 
 	// run a container with the image
-	out, _ = dockerCmd(c, "run", "-d", "busybox-one", "top")
+	out, _ = runSleepingContainerInImage(c, "busybox-one")
 
 	containerID = strings.TrimSpace(out)
 
 	// first checkout without force it fails
-	out, _, err = dockerCmdWithError("rmi", imgID)
+	out, _, err := dockerCmdWithError("rmi", imgID)
 	expected := fmt.Sprintf("conflict: unable to delete %s (cannot be forced) - image is being used by running container %s", stringid.TruncateID(imgID), stringid.TruncateID(containerID))
 	// rmi tagged in multiple repos should have failed without force
 	c.Assert(err, checker.NotNil)
@@ -100,10 +105,17 @@ func (s *DockerSuite) TestRmiImgIDMultipleTag(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiImgIDForce(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	out, _ := dockerCmd(c, "run", "-d", "busybox", "/bin/sh", "-c", "mkdir '/busybox-test'")
 
 	containerID := strings.TrimSpace(out)
+
+	// Wait for it to exit as cannot commit a running container on Windows, and
+	// it will take a few seconds to exit
+	if daemonPlatform == "windows" {
+		err := waitExited(containerID, 60*time.Second)
+		c.Assert(err, check.IsNil)
+	}
+
 	dockerCmd(c, "commit", containerID, "busybox-test")
 
 	imagesBefore, _ := dockerCmd(c, "images", "-a")
@@ -115,11 +127,10 @@ func (s *DockerSuite) TestRmiImgIDForce(c *check.C) {
 		imagesAfter, _ := dockerCmd(c, "images", "-a")
 		c.Assert(strings.Count(imagesAfter, "\n"), checker.Equals, strings.Count(imagesBefore, "\n")+4, check.Commentf("before: %q\n\nafter: %q\n", imagesBefore, imagesAfter))
 	}
-	imgID, err := inspectField("busybox-test", "Id")
-	c.Assert(err, checker.IsNil)
+	imgID := inspectField(c, "busybox-test", "Id")
 
 	// first checkout without force it fails
-	out, _, err = dockerCmdWithError("rmi", imgID)
+	out, _, err := dockerCmdWithError("rmi", imgID)
 	// rmi tagged in multiple repos should have failed without force
 	c.Assert(err, checker.NotNil)
 	// rmi tagged in multiple repos should have failed without force
@@ -135,14 +146,13 @@ func (s *DockerSuite) TestRmiImgIDForce(c *check.C) {
 
 // See https://github.com/docker/docker/issues/14116
 func (s *DockerSuite) TestRmiImageIDForceWithRunningContainersAndMultipleTags(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	dockerfile := "FROM busybox\nRUN echo test 14116\n"
 	imgID, err := buildImage("test-14116", dockerfile, false)
 	c.Assert(err, checker.IsNil)
 
 	newTag := "newtag"
 	dockerCmd(c, "tag", imgID, newTag)
-	dockerCmd(c, "run", "-d", imgID, "top")
+	runSleepingContainerInImage(c, imgID)
 
 	out, _, err := dockerCmdWithError("rmi", "-f", imgID)
 	// rmi -f should not delete image with running containers
@@ -151,7 +161,6 @@ func (s *DockerSuite) TestRmiImageIDForceWithRunningContainersAndMultipleTags(c
 }
 
 func (s *DockerSuite) TestRmiTagWithExistingContainers(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	container := "test-delete-tag"
 	newtag := "busybox:newtag"
 	bb := "busybox:latest"
@@ -164,7 +173,6 @@ func (s *DockerSuite) TestRmiTagWithExistingContainers(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiForceWithExistingContainers(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	image := "busybox-clone"
 
 	cmd := exec.Command(dockerBinary, "build", "--no-cache", "-t", image, "-")
@@ -180,13 +188,12 @@ MAINTAINER foo`)
 }
 
 func (s *DockerSuite) TestRmiWithMultipleRepositories(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	newRepo := "127.0.0.1:5000/busybox"
 	oldRepo := "busybox"
 	newTag := "busybox:test"
 	dockerCmd(c, "tag", oldRepo, newRepo)
 
-	dockerCmd(c, "run", "--name", "test", oldRepo, "touch", "/home/abcd")
+	dockerCmd(c, "run", "--name", "test", oldRepo, "touch", "/abcd")
 
 	dockerCmd(c, "commit", "test", newTag)
 
@@ -195,13 +202,12 @@ func (s *DockerSuite) TestRmiWithMultipleRepositories(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiForceWithMultipleRepositories(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	imageName := "rmiimage"
 	tag1 := imageName + ":tag1"
 	tag2 := imageName + ":tag2"
 
 	_, err := buildImage(tag1,
-		`FROM scratch
+		`FROM busybox
 		MAINTAINER "docker"`,
 		true)
 	if err != nil {
@@ -220,25 +226,16 @@ func (s *DockerSuite) TestRmiForceWithMultipleRepositories(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiBlank(c *check.C) {
-	testRequires(c, DaemonIsLinux)
-	// try to delete a blank image name
-	out, _, err := dockerCmdWithError("rmi", "")
-	// Should have failed to delete '' image
+	out, _, err := dockerCmdWithError("rmi", " ")
+	// Should have failed to delete ' ' image
 	c.Assert(err, checker.NotNil)
 	// Wrong error message generated
 	c.Assert(out, checker.Not(checker.Contains), "no such id", check.Commentf("out: %s", out))
 	// Expected error message not generated
 	c.Assert(out, checker.Contains, "image name cannot be blank", check.Commentf("out: %s", out))
-
-	out, _, err = dockerCmdWithError("rmi", " ")
-	// Should have failed to delete ' ' image
-	c.Assert(err, checker.NotNil)
-	// Expected error message not generated
-	c.Assert(out, checker.Contains, "image name cannot be blank", check.Commentf("out: %s", out))
 }
 
 func (s *DockerSuite) TestRmiContainerImageNotFound(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	// Build 2 images for testing.
 	imageNames := []string{"test1", "test2"}
 	imageIds := make([]string, 2)
@@ -250,7 +247,7 @@ func (s *DockerSuite) TestRmiContainerImageNotFound(c *check.C) {
 	}
 
 	// Create a long-running container.
-	dockerCmd(c, "run", "-d", imageNames[0], "top")
+	runSleepingContainerInImage(c, imageNames[0])
 
 	// Create a stopped container, and then force remove its image.
 	dockerCmd(c, "run", imageNames[1], "true")
@@ -265,9 +262,8 @@ func (s *DockerSuite) TestRmiContainerImageNotFound(c *check.C) {
 
 // #13422
 func (s *DockerSuite) TestRmiUntagHistoryLayer(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	image := "tmp1"
-	// Build a image for testing.
+	// Build an image for testing.
 	dockerfile := `FROM busybox
 MAINTAINER foo
 RUN echo 0 #layer0
@@ -312,10 +308,7 @@ RUN echo 2 #layer2
 }
 
 func (*DockerSuite) TestRmiParentImageFail(c *check.C) {
-	testRequires(c, DaemonIsLinux)
-
-	parent, err := inspectField("busybox", "Parent")
-	c.Assert(err, check.IsNil)
+	parent := inspectField(c, "busybox", "Parent")
 	out, _, err := dockerCmdWithError("rmi", parent)
 	c.Assert(err, check.NotNil)
 	if !strings.Contains(out, "image has dependent child images") {
@@ -324,14 +317,15 @@ func (*DockerSuite) TestRmiParentImageFail(c *check.C) {
 }
 
 func (s *DockerSuite) TestRmiWithParentInUse(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	out, _ := dockerCmd(c, "create", "busybox")
 	cID := strings.TrimSpace(out)
+
 	out, _ = dockerCmd(c, "commit", cID)
 	imageID := strings.TrimSpace(out)
 
 	out, _ = dockerCmd(c, "create", imageID)
 	cID = strings.TrimSpace(out)
+
 	out, _ = dockerCmd(c, "commit", cID)
 	imageID = strings.TrimSpace(out)
 
@@ -340,17 +334,14 @@ func (s *DockerSuite) TestRmiWithParentInUse(c *check.C) {
 
 // #18873
 func (s *DockerSuite) TestRmiByIDHardConflict(c *check.C) {
-	testRequires(c, DaemonIsLinux)
 	dockerCmd(c, "create", "busybox")
 
-	imgID, err := inspectField("busybox:latest", "Id")
-	c.Assert(err, checker.IsNil)
+	imgID := inspectField(c, "busybox:latest", "Id")
 
-	_, _, err = dockerCmdWithError("rmi", imgID[:12])
+	_, _, err := dockerCmdWithError("rmi", imgID[:12])
 	c.Assert(err, checker.NotNil)
 
 	// check that tag was not removed
-	imgID2, err := inspectField("busybox:latest", "Id")
-	c.Assert(err, checker.IsNil)
+	imgID2 := inspectField(c, "busybox:latest", "Id")
 	c.Assert(imgID, checker.Equals, imgID2)
 }
